import hashlib
from binascii import hexlify


def sha256(payload: bytes) -> bytes:
    return hashlib.sha256(payload).digest()


def double_sha256(payload: bytes) -> bytes:
    return sha256(sha256(payload))


def double_sha256_checksum(payload: bytes) -> bytes:
    return double_sha256(payload)[:4]


def ripemd160_sha256(payload: bytes) -> bytes:
    return hashlib.new('ripemd160', sha256(payload)).digest()


BASE58_ALPHABET = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'


def b58_encode(payload: bytes) -> str:
    pad = 0
    for byte in payload:
        if byte == 0:
            pad += 1
        else:
            break
    prefix = '1' * pad
    num = int.from_bytes(payload, 'big')
    result = ''
    while num > 0:
        num, remaining = divmod(num, 58)
        result = BASE58_ALPHABET[remaining] + result
    return prefix + result


def b58check_encode(payload: bytes) -> str:
    return b58_encode(payload + double_sha256_checksum(payload))


def int_to_bytes(num: int, byteorder: str = 'big') -> bytes:
    """Converts an int to the least number of bytes as possible."""
    return num.to_bytes((num.bit_length() + 7) // 8, byteorder)


def b58_decode(encoded: str) -> bytes:
    pad = 0
    for char in encoded:
        if char == '1':
            pad += 1
        else:
            break
    prefix = b'\x00' * pad
    num = 0
    try:
        for char in encoded:
            num *= 58
            num += BASE58_ALPHABET.index(char)
    except KeyError:
        raise ValueError(f'Invalid base58 encoded "{encoded}"')
    return prefix + int_to_bytes(num)


def b58check_decode(encoded: str) -> bytes:
    decoded = b58_decode(encoded)
    payload = decoded[:-4]
    decoded_checksum = decoded[-4:]
    hash_checksum = double_sha256_checksum(payload)
    if decoded_checksum != hash_checksum:
        raise ValueError(f'Decoded checksum {hexlify(decoded_checksum)} from "{hexlify(decoded)}" is not equal to hash checksum {hexlify(hash_checksum)}')
    return payload


from bsvpy.ec_point_operation import curve
import random


class Polynomial:
    """
    Polynomial y = a0 * x^0 + a1 * x^1 + ... + at * x^t on finite field Secp256k1.n
    """

    @staticmethod
    def random(order: int, debug: bool = False) -> 'Polynomial':
        """Random a polynomial with the specific order"""
        if order < 1:
            raise ValueError(f'The polynomial order should be a positive integer.')
        range_stop = 10 if debug else curve.n
        coefficients = [random.randrange(1, range_stop)]
        for i in range(order):
            coefficients.append(random.randrange(1, range_stop))
        return Polynomial(coefficients)

    @staticmethod
    def interpolate_evaluate(points: list, x: int) -> int:
        """Lagrange interpolate with the giving points, then evaluate y at x"""
        if len(points) < 2:
            raise ValueError('Lagrange interpolation requires at least 2 points')
        # [(numerator, denominator) ...]
        lagrange = [(0, 0)] * len(points)
        # the product of all the denominator
        denominator_product = 1
        for i in range(len(points)):
            numerator, denominator = 1, 1
            for j in range(len(points)):
                if j != i:
                    numerator *= (x - points[j][0])
                    denominator *= (points[i][0] - points[j][0])
            lagrange[i] = (points[i][1] * numerator, denominator)
            denominator_product *= denominator
        numerator_sum = 0
        for (numerator, denominator) in lagrange:
            numerator_sum += numerator * denominator_product // denominator
        return (numerator_sum // denominator_product) % curve.n

    def __init__(self, coefficients: list) -> None:
        if len(coefficients) < 2:
            raise ValueError(f'The polynomial should have 2 coefficients at least.')
        self.order = len(coefficients) - 1
        self.coefficients = coefficients[:]

    def evaluate(self, x: int) -> int:
        """Calculate y for the specific x"""
        if x == 0:
            return self.coefficients[0]
        y, pow_x = 0, 1
        for i in range(len(self.coefficients)):
            y += self.coefficients[i] * pow_x
            pow_x *= x
        return y % curve.n

    def add(self, other: 'Polynomial') -> 'Polynomial':
        """Returns the polynomial = self + other"""
        coefficients = []
        i = 0
        while i < len(self.coefficients) and i < len(other.coefficients):
            coefficients.append(self.coefficients[i] + other.coefficients[i])
            i += 1
        while i < len(self.coefficients):
            coefficients.append(self.coefficients[i])
            i += 1
        while i < len(other.coefficients):
            coefficients.append(other.coefficients[i])
            i += 1
        for i in range(len(coefficients)):
            coefficients[i] %= curve.n
        return Polynomial(coefficients)

    def multiply(self, other: 'Polynomial') -> 'Polynomial':
        """Returns the polynomial = self * other"""
        coefficients = [0] * (self.order + other.order + 1)
        for i in range(len(self.coefficients)):
            for j in range(len(other.coefficients)):
                coefficients[i + j] += self.coefficients[i] * other.coefficients[j]
        for i in range(len(coefficients)):
            coefficients[i] %= curve.n
        return Polynomial(coefficients)

    def __str__(self) -> str:
        return f'<Polynomial order={self.order}, coefficients=[{", ".join(str(i) for i in self.coefficients)}]>'

    def __eq__(self, other: 'Polynomial') -> bool:
        return self.coefficients == other.coefficients


if __name__ == '__main__':
    a = 2

    f = Polynomial.random(order=3, debug=True)
    print(f'f = {f}')
    fa = f.evaluate(a)
    print(f'f({a}) = {fa}')

    g = Polynomial.random(order=4, debug=True)
    print(f'g = {g}')
    ga = g.evaluate(a)
    print(f'g({a}) = {ga}')

    v = f.add(g)
    assert v == g.add(f)
    print(f'v = f + g = {v}')
    va = v.evaluate(a)
    print(f'v({a}) = {va}')
    assert va == (fa + ga) % curve.n

    u = f.multiply(g)
    assert u == g.multiply(f)
    print(f'u = f * g = {u}')
    ua = u.evaluate(a)
    print(f'u({a}) = {ua}')
    assert ua == (fa * ga) % curve.n

    p = [(1, 350), (2, 770), (3, 1350)]
    assert Polynomial.interpolate_evaluate(p, 0) == 90
    assert Polynomial.interpolate_evaluate(p, 1) == 350
    assert Polynomial.interpolate_evaluate(p, 2) == 770



from bsvpy.meta import int_to_varint, public_key_to_address, address_to_public_key_hash, public_key_hash
from bsvpy.modular_inverse import modular_multiplicative_inverse
from bsvpy.ec_point_operation import curve, add, scalar_multiply
from bsvpy.sign import hash_to_int, sign_recoverable, verify_signature
from base64 import b64encode, b64decode


def message_bytes(message: str) -> bytes:
    """Serialize plain text message to format (LEN || message.utf-8)"""
    msg_bytes = message.encode('utf-8')
    return int_to_varint(len(msg_bytes)) + msg_bytes


def message_digest(message: str) -> bytes:
    """Returns the digest of plain text message"""
    return message_bytes('Bitcoin Signed Message:\n') + message_bytes(message)


def sign_message(private_key: int, plain_text: str) -> tuple:
    """Sign arbitrary message with bitcoin private key, returns (p2pkh_address, serialized_compact_signature)"""
    d = message_digest(plain_text)
    # recovery signature
    recovery_id, r, s = sign_recoverable(private_key, d)
    # p2pkh address
    public_key = scalar_multiply(private_key, curve.g)
    p2pkh_address = public_key_to_address(public_key, compressed=True)
    # prefix = 27 + recovery_id + (4 if using compressed public key else 0)
    prefix = 27 + recovery_id + 4
    serialized_sig = prefix.to_bytes(1, byteorder='big') + r.to_bytes(32, byteorder='big') + s.to_bytes(32, byteorder='big')
    return p2pkh_address, b64encode(serialized_sig).decode('ascii')


def verify_message(p2pkh_address: str, plain_text: str, signature: str) -> bool:
    """Verify serialized compact signature with p2pkh address and plain text"""
    sig_bytes = b64decode(signature)
    if len(sig_bytes) != 65:
        return False
    prefix, r, s = sig_bytes[0], int.from_bytes(sig_bytes[1:33], byteorder='big'), int.from_bytes(sig_bytes[33:], byteorder='big')
    # Calculate recovery_id
    compressed = False
    if prefix < 27 or prefix >= 35:
        return False
    if prefix >= 31:
        compressed = True
        prefix -= 4
    recovery_id = prefix - 27
    # Recover point kG, k is the ephemeral private key
    x = r + (curve.n if recovery_id >= 2 else 0)
    y_squared = (x * x * x + curve.a * x + curve.b) % curve.p
    y = pow(y_squared, (curve.p + 1) // 4, curve.p)
    if (y + recovery_id) % 2 != 0:
        y = -y % curve.p
    point_k = (x, y)
    # Calculate point aG, a is the private key
    d = message_digest(plain_text)
    e = hash_to_int(d)
    mod_inv_r = modular_multiplicative_inverse(r, curve.n)
    public_key = add(scalar_multiply(mod_inv_r * s, point_k), scalar_multiply(mod_inv_r * (-e % curve.n), curve.g))
    # Verify signature
    if not verify_signature(public_key, d, (r, s)):
        return False
    # Check public key hash
    if public_key_hash(public_key, compressed) != address_to_public_key_hash(p2pkh_address):
        return False
    # OK
    return True


if __name__ == '__main__':
    priv_key = 0xf97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62
    pub_key = scalar_multiply(priv_key, curve.g)
    plain = '使用比特币私钥对任意消息签名 aaron67'
    print(plain)
    # Create signature
    print(sign_message(priv_key, plain))
    # Verify signature created in ElectrumSV
    address = public_key_to_address(pub_key, compressed=True)
    sig_electrum = 'IGdzMq98lowek10e3JFXWj909xp0oLRj71aF7jpWRxaabwH+fBia/K2JpoGQlFFbAl/Q5jo2DYSzQw6pZWhmRtk='
    print(verify_message(address, plain, sig_electrum))



from bsvpy.ec_point_operation import curve
from bsvpy.crypto import ripemd160_sha256, b58check_encode, b58check_decode, sha256, b58_encode
from binascii import hexlify


def int_to_varint(value: int) -> bytes:
    if value <= 0xfc:
        return value.to_bytes(1, 'little')
    elif value <= 0xffff:
        return b'\xfd' + value.to_bytes(2, 'little')
    elif value <= 0xffffffff:
        return b'\xfe' + value.to_bytes(4, 'little')
    else:
        return b'\xff' + value.to_bytes(8, 'little')


def serialize_public_key(public_key: tuple, compressed: bool = True) -> bytes:
    """Serialize public key point to compressed format (02 || x) or (03 || x), or uncompressed format (04 || x || y)"""
    x, y = public_key
    if compressed:
        return (b'\x02' if y % 2 == 0 else b'\x03') + x.to_bytes(32, byteorder='big')
    return b'\x04' + x.to_bytes(32, byteorder='big') + y.to_bytes(32, byteorder='big')


def public_key_hash(public_key: tuple, compressed: bool = True) -> bytes:
    public_key_bytes = serialize_public_key(public_key, compressed)
    return ripemd160_sha256(public_key_bytes)


def public_key_to_address(public_key: tuple, compressed: bool = True) -> str:
    return b58check_encode(b'\x00' + public_key_hash(public_key, compressed))


def address_to_public_key_hash(address: str) -> bytes:
    """Decode p2pkh address to public key hash"""
    decoded = b58check_decode(address)
    assert decoded[:1] == b'\x00'
    return decoded[1:]


OP_DUP = b'\x76'
OP_HASH160 = b'\xa9'
OP_PUSH_20 = b'\x14'
OP_EQUALVERIFY = b'\x88'
OP_CHECKSIG = b'\xac'


def build_locking_script(pkh: bytes) -> bytes:
    """Build p2pkh locking script for the given public key hash"""
    script = OP_DUP + OP_HASH160 + OP_PUSH_20 + pkh + OP_EQUALVERIFY + OP_CHECKSIG
    return int_to_varint(len(script)) + script


def serialize_signature(signature: tuple) -> bytes:
    """Serialize ECDSA signature (r, s) to bitcoin strict DER format."""
    r, s = signature
    # BIP-62, BIP-66
    # Enforce low s value in signature
    # Using (curve.n - s) if s > 0x7fffffffffffffffffffffffffffffff5d576e7357a4501ddfe92f46681b20a0
    # https://en.bitcoin.it/wiki/Transaction_malleability
    if s > curve.n // 2:
        s = curve.n - s
    # r
    r_bytes = r.to_bytes(32, byteorder='big').lstrip(b'\x00')
    if r_bytes[0] & 0x80:
        r_bytes = b'\x00' + r_bytes
    serialized = bytes([2, len(r_bytes)]) + r_bytes
    # s
    s_bytes = s.to_bytes(32, byteorder='big').lstrip(b'\x00')
    if s_bytes[0] & 0x80:
        s_bytes = b'\x00' + s_bytes
    serialized += bytes([2, len(s_bytes)]) + s_bytes
    return bytes([0x30, len(serialized)]) + serialized


def deserialize_signature(serialized: bytes) -> tuple:
    """Deserialize ECDSA bitcoin DER formatted signature to (r, s)"""
    try:
        assert serialized[0] == 0x30
        assert int(serialized[1]) == len(serialized) - 2
        # r
        assert serialized[2] == 0x02
        r_len = int(serialized[3])
        r = int.from_bytes(serialized[4: 4 + r_len], byteorder='big')
        # s
        assert serialized[4 + r_len] == 0x02
        s_len = int(serialized[5 + r_len])
        s = int.from_bytes(serialized[-s_len:], byteorder='big')
        return r, s
    except Exception:
        raise ValueError(f'Invalid DER encoded {hexlify(serialized)}.')


def private_key_to_wif(private_key: int, compressed: bool = True) -> str:
    payload = b'\x80' + private_key.to_bytes(32, byteorder='big')
    if compressed:
        payload += b'\x01'
    checksum = sha256(sha256(payload))[0:4]
    return b58_encode(payload + checksum)


def wif_to_private_key(wif: str) -> int:
    if not wif.startswith('5') and not wif.startswith('K') and not wif.startswith('L'):
        raise ValueError(f'Invalid WIF {wif}')
    payload = b58check_decode(wif)
    if wif.startswith('K') or wif.startswith('L'):
        payload = payload[:-1]
    return int.from_bytes(payload[1:], byteorder='big')


if __name__ == '__main__':
    sig = (114587593887127314608220924841831336233967095853165151956820984900193959037698, 24000727837347392504013031837120627225728348681623127776947626422811445180558)
    serialized_sig = serialize_signature(sig)
    print(hexlify(serialized_sig))
    decoded_sig = deserialize_signature(serialized_sig)
    print(decoded_sig == sig)



import collections
from bsvpy.modular_inverse import modular_multiplicative_inverse

EllipticCurve = collections.namedtuple('EllipticCurve', 'name p a b g n h')

curve = EllipticCurve(
    name='Secp256k1',
    p=0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffefffffc2f,
    a=0,
    b=7,
    g=(0x79be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798, 0x483ada7726a3c4655da4fbfc0e1108a8fd17b448a68554199c47d08ffb10d4b8),
    n=0xfffffffffffffffffffffffffffffffebaaedce6af48a03bbfd25e8cd0364141,
    h=1,
)


def on_curve(point: tuple) -> bool:
    """Returns True if the given point lies on the elliptic curve."""
    if point is None:
        # None represents the point at infinity.
        return True
    x, y = point
    return (y * y - x * x * x - curve.a * x - curve.b) % curve.p == 0


def negative(point: tuple) -> tuple or None:
    """Returns -point."""
    assert on_curve(point)
    if point is None:
        # -0 = 0
        return None
    x, y = point
    result = (x, -y % curve.p)
    assert on_curve(result)
    return result


def add(p: tuple, q: tuple) -> tuple or None:
    """Returns the result of p + q according to the group law."""
    assert on_curve(p)
    assert on_curve(q)
    if p is None:
        # 0 + q = q
        return q
    if q is None:
        # p + 0 = p
        return p
    #
    # p == -q
    #
    if p == negative(q):
        return None
    #
    # p != -q
    #
    x1, y1 = p
    x2, y2 = q
    if p == q:
        m = (3 * x1 * x1 + curve.a) * modular_multiplicative_inverse(2 * y1, curve.p)
    else:
        m = (y1 - y2) * modular_multiplicative_inverse(x1 - x2, curve.p)
    x = m * m - x1 - x2
    y = y1 + m * (x - x1)
    result = (x % curve.p, -y % curve.p)
    assert on_curve(result)
    return result


def scalar_multiply(k: int, point: tuple) -> tuple or None:
    """Returns k * point computed using the double and add algorithm."""
    assert on_curve(point)
    if k % curve.n == 0 or point is None:
        return None
    if k < 0:
        # k * point = -k * (-point)
        return scalar_multiply(-k, negative(point))
    # Double and add
    result = None
    while k:
        if k & 1:
            result = add(result, point)
        point = add(point, point)
        k >>= 1
    assert on_curve(result)
    return result


if __name__ == '__main__':
    a = 0xf97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62
    print('a =', hex(a))
    ag = scalar_multiply(a, curve.g)
    print('x =', hex(ag[0]))
    print('y =', hex(ag[1]))

    from meta import public_key_to_address, private_key_to_wif

    print(private_key_to_wif(a))
    print(public_key_to_address(ag))



from bsvpy.ec_point_operation import curve, scalar_multiply
from bsvpy.meta import int_to_varint, address_to_public_key_hash, build_locking_script, deserialize_signature, serialize_signature, serialize_public_key, public_key_to_address, private_key_to_wif
from bsvpy.crypto import double_sha256
from bsvpy.sign import verify_signature, sign
from collections import namedtuple
from binascii import unhexlify, hexlify
from urllib import request
import json
import requests

VERSION = 0x01.to_bytes(4, 'little')
SEQUENCE = 0xffffffff.to_bytes(4, byteorder='little')
LOCK_TIME = 0x00.to_bytes(4, byteorder='little')

SH_ALL = 0x01
SH_FORKID = 0x40
SIGHASH_ALL = SH_ALL | SH_FORKID
P_0 = b'\x00'
OP_FALSE = b'\00'
OP_CHECKLOCKTIMEVERIFY = b'\xb1'
OP_CHECKSIG = b'\xac'
OP_DUP = b'v'
OP_EQUALVERIFY = b'\x88'
OP_HASH160 = b'\xa9'
OP_PUSH_20 = b'\x14'
OP_RETURN = b'\x6a'
OP_PUSHDATA1 = b'\x4c'
OP_PUSHDATA2 = b'\x4d'
OP_PUSHDATA4 = b'\x4e'




OP_DUP = b'\x76'
OP_HASH160 = b'\xa9'
OP_PUSH_20 = b'\x14'
OP_EQUALVERIFY = b'\x88'
OP_CHECKSIG = b'\xac'


class TxIn:
    def __init__(self, satoshi: int, txid: str, index: int, locking_script: str, sequence: bytes = SEQUENCE) -> None:
        self.satoshi = satoshi.to_bytes(8, byteorder='little')
        self.txid = unhexlify(txid)[::-1]
        self.index = index.to_bytes(4, byteorder='little')
        self.locking_script = unhexlify(locking_script)
        self.locking_script_len = int_to_varint(len(self.locking_script))
        self.unlocking_script = b''
        self.unlocking_script_len = b''
        self.sequence = sequence


#TxOut = namedtuple('TxOut', 'address satoshi')

def get_op_pushdata_code(dest):
    length_data = len(dest)
    if length_data <= 0x4c:  # (https://en.bitcoin.it/wiki/Script)
        return length_data.to_bytes(1, byteorder='little')
    elif length_data <= 0xff:
        return OP_PUSHDATA1 + length_data.to_bytes(1, byteorder='little')  # OP_PUSHDATA1 format
    elif length_data <= 0xffff:
        return OP_PUSHDATA2 + length_data.to_bytes(2, byteorder='little')  # OP_PUSHDATA2 format
    else:
        return OP_PUSHDATA4 + length_data.to_bytes(4, byteorder='little')  # OP_PUSHDATA4 format


def serialize_outputs(outputs: list,custom_pushdata=False) -> bytes:

    """
    Serialize outputs [(address, satoshi), (address, satoshi), ...]
    to format (satoshi || LEN(locking_script) || locking_script) || (satoshi || LEN(locking_script) || locking_script) || ...)
    """
    output_block = b''
    #print(outputs)
    for data in outputs:
        dest, amount = data
        #print('data:',data)
        #print('dest:',dest)
        #print('amount:',amount)

        # Real recipient
        if amount:
            script = (OP_DUP + OP_HASH160 + OP_PUSH_20 +
                      address_to_public_key_hash(dest) +
                      OP_EQUALVERIFY + OP_CHECKSIG)
            #print(address_to_public_key_hash(dest))

            output_block += amount.to_bytes(8, byteorder='little')

        # Blockchain storage
        else:
            if custom_pushdata is False:
                script = OP_FALSE + OP_RETURN + get_op_pushdata_code(dest) + dest
                output_block += b'\x00\x00\x00\x00\x00\x00\x00\x00'

            elif custom_pushdata is True:
                #manual control over number of bytes in each batch of pushdata
                if type(dest) != bytes:
                    raise TypeError("custom pushdata must be of type: bytes")
                else:
                    script = (OP_FALSE + OP_RETURN + dest)

                output_block += b'\x00\x00\x00\x00\x00\x00\x00\x00'

        # Script length in wiki is "Var_int" but there's a note of "modern BitcoinQT" using a more compact "CVarInt"
        output_block += int_to_varint(len(script))
        output_block += script
        output_bytes=output_block

        

    #return output_block

    #print('output_bytes1:',output_bytes) 
    #output_bytes = b''
    #for output in outputs:
        #output_bytes += output.satoshi.to_bytes(8, byteorder='little') + build_locking_script(address_to_public_key_hash(output.address))
    #print('output_bytes2:',output_bytes)   
    return output_bytes


def transaction_digest(tx_ins: list, tx_outs: list, lock_time: bytes = LOCK_TIME, sighash: int = SIGHASH_ALL) -> list:
    """Returns the digest of unsigned transaction according to SIGHASH"""
    # BIP-143 https://github.com/bitcoin/bips/blob/master/bip-0143.mediawiki
    #  1. nVersion of the transaction (4-byte little endian)
    #  2. hashPrevouts (32-byte hash)
    #  3. hashSequence (32-byte hash)
    #  4. outpoint (32-byte hash + 4-byte little endian)
    #  5. scriptCode of the input (serialized as scripts inside CTxOuts)
    #  6. value of the output spent by this input (8-byte little endian)
    #  7. nSequence of the input (4-byte little endian)
    #  8. hashOutputs (32-byte hash)
    #  9. nLocktime of the transaction (4-byte little endian)
    # 10. sighash type of the signature (4-byte little endian)
    if sighash == SIGHASH_ALL:
        hash_prevouts = double_sha256(b''.join([tx_in.txid + tx_in.index for tx_in in tx_ins]))
        hash_sequence = double_sha256(b''.join([tx_in.sequence for tx_in in tx_ins]))
        hash_outputs = double_sha256(serialize_outputs(tx_outs))
        digests = []
        for tx_in in tx_ins:
            digests.append(
                VERSION +
                hash_prevouts + hash_sequence +
                tx_in.txid + tx_in.index + tx_in.locking_script_len + tx_in.locking_script + tx_in.satoshi + tx_in.sequence +
                hash_outputs +
                lock_time +
                sighash.to_bytes(4, byteorder='little')
            )
        return digests
    raise ValueError(f'Unsupported SIGHASH value {sighash}')


def serialize_transaction(tx_ins: list, tx_outs: list, lock_time: bytes = LOCK_TIME) -> bytes:
    """Serialize signed transaction"""
    # version
    raw_transaction = VERSION
    # inputs
    raw_transaction += int_to_varint(len(tx_ins))
    for tx_in in tx_ins:
        raw_transaction += tx_in.txid + tx_in.index + tx_in.unlocking_script_len + tx_in.unlocking_script + tx_in.sequence
    # outputs
    raw_transaction += int_to_varint(len(tx_outs)) + serialize_outputs(tx_outs)
    # lock_time
    raw_transaction += lock_time
    return raw_transaction



def address_str(priv_key:int):#整数key求地址
    
    pub_key = scalar_multiply(priv_key, curve.g)
    return public_key_to_address(pub_key)
    
def pub_key(priv_key:int):#整数key求公钥
    
    return scalar_multiply(priv_key, curve.g)

def get_data(url):# 获取网页数据

    req = request.Request(url)
    with request.urlopen(req) as f:
        data=f.read().decode()
    return data


    
def scriptPubKey(address):#获取锁定脚本
    url='https://api.whatsonchain.com/v1/bsv/main/address/{0}/info'.format(address)
    a=get_data(url).replace('true','"true"')
    b=a.replace('false','"false"')

    return(eval(b)["scriptPubKey"])
    
    

#print(address(123))
#print(pub_key(123))


def send_to(priv_key:int,send_to_address:str,send_to_satoshi=160,send_to_message=None):
    
    pub_key = scalar_multiply(priv_key, curve.g) # 公钥

    address=address_str(priv_key) #发送方地址
    
    
    url='https://api.whatsonchain.com/v1/bsv/main/address/{0}/unspent'.format(address)
    unspent=eval(get_data(url))
    address_locking=scriptPubKey(address)
    inputssss=''
    all_satoshi=0
    for i in unspent:
        unspent_satoshi=i['value']
        unspent_txid=i['tx_hash']
        unspent_index=i['tx_pos']
        all_satoshi+=unspent_satoshi  #未花费总数
        inputssss=inputssss+'TxIn(satoshi={0}, txid="{1}", index={2}, locking_script="{3}"),'.format(unspent_satoshi,unspent_txid,unspent_index,address_locking)
    inputss='['+inputssss[:-1]+']'
    #print(inputss)
    inputs=eval(inputss)
    #print(inputs)
        
    

    
    #inputs = [
        #TxIn(satoshi=100000, txid='b811bc9bbcc84c6b3a2376511478ee7d262e8ab6809b775f115cbce0577841d9', index=0, locking_script='76a914894acc947768ee6f16929e47ec4a0af137da086088ac'),]
    #print(inputs)
     

    serialized_pub_key = serialize_public_key(pub_key)
    tx_inputs = inputs[0:]
    #tx_outputs = [TxOut(address='1Lkqjn1L8Th2PkojGTorASNJL8s1mj1xZR', satoshi=111),]

    send_to_message=send_to_message.encode('utf-8')
    len_message=len(send_to_message) # 字节长度
    self_satoshi=all_satoshi-send_to_satoshi- len_message-200   #剩余返回自己
    
    tx_outputs = [(send_to_address,send_to_satoshi),(send_to_message,0),(address,self_satoshi)] #剩余返回自己

    
    #print(tx_outputs)
    tx_digests = transaction_digest(tx_inputs, tx_outputs)
    for i in range(len(tx_digests)):
        tx_digest = tx_digests[i]
        sig = sign(priv_key, tx_digest)
        serialized_sig = serialize_signature(sig)
        # unlocking_script = LEN + der + sighash + LEN + public_key
        tx_inputs[i].unlocking_script = bytes([len(serialized_sig) + 1]) + serialized_sig + bytes([SIGHASH_ALL, len(serialized_pub_key)]) + serialized_pub_key
        #print(hexlify(tx_inputs[i].unlocking_script))
        tx_inputs[i].unlocking_script_len = int_to_varint(len(tx_inputs[i].unlocking_script))
        #print(hexlify(tx_inputs[i].unlocking_script_len))
    raw = serialize_transaction(tx_inputs, tx_outputs)
    
    raw_tx=hexlify(raw).decode()
    #print(raw_tx)
    url='https://api.whatsonchain.com/v1/bsv/main/tx/raw'
    data = json.dumps({"txhex":raw_tx})
    r =requests.post(url, data=data, timeout=10)
    
    return r.json() # 返回交易哈希或者错误信息


    #print(get_data(url))
    #return hexlify(raw)
    #tx_id = double_sha256(raw)[::-1]
    #print(hexlify(tx_id))

if __name__ == '__main__':
    send_to(priv_key=123,send_to_address='1Lkqjn1L8Th2PkojGTorASNJL8s1mj1xZR',send_to_message='80tes1111t')




from bsvpy.ec_point_operation import curve, add, scalar_multiply
from bsvpy.modular_inverse import modular_multiplicative_inverse
from bsvpy.crypto import double_sha256, sha256
import random


def hash_to_int(message: bytes) -> int:
    """Calculate the bitcoin double-sha256 hash of the message, return as an integer"""
    h = double_sha256(message)
    return int.from_bytes(h, byteorder='big')


def sign(private_key: int, message: bytes) -> tuple:
    """Create ECDSA signature (r, s)"""
    e = hash_to_int(message)
    r, s = 0, 0
    while not r or not s:
        k = random.randrange(1, curve.n)
        k_x, _ = scalar_multiply(k, curve.g)
        r = k_x % curve.n
        s = ((e + r * private_key) * modular_multiplicative_inverse(k, curve.n)) % curve.n
    return r, s


def sign_recoverable(private_key: int, message: bytes) -> tuple:
    """Create recoverable ECDSA signature, aka compact signature, (recovery_id, r, s)"""
    e = hash_to_int(message)
    recovery_id, r, s = 0, 0, 0
    while not r or not s:
        k = random.randrange(1, curve.n)
        k_x, k_y = scalar_multiply(k, curve.g)
        # r
        r = k_x % curve.n
        recovery_id = 0 | 2 if k_x > curve.n else 0 | k_y % 2
        # s
        s = ((e + r * private_key) * modular_multiplicative_inverse(k, curve.n)) % curve.n
    return recovery_id, r, s


def verify_signature(public_key: tuple, message: bytes, signature: tuple) -> bool:
    """Verify signature with public key and message"""
    e = hash_to_int(message)
    r, s = signature
    w = modular_multiplicative_inverse(s, curve.n)
    u1 = (w * e) % curve.n
    u2 = (w * r) % curve.n
    x, _ = add(scalar_multiply(u1, curve.g), scalar_multiply(u2, public_key))
    return r == (x % curve.n)


if __name__ == '__main__':
    priv_key = 0xf97c89aaacf0cd2e47ddbacc97dae1f88bec49106ac37716c451dcdd008a4b62
    pub_key = scalar_multiply(priv_key, curve.g)
    plain_text = '你好世界'
    digest = sha256(plain_text.encode('utf-8'))
    # Sign
    sig_r, sig_s = sign(priv_key, digest)
    print(' r =', sig_r)
    print(' s =', sig_s)
    # Verify (r, s)
    print(verify_signature(pub_key, digest, (sig_r, sig_s)))
    # Verify (r, -s)
    negative_s = -sig_s % curve.n
    print('-s =', negative_s)
    print(verify_signature(pub_key, digest, (sig_r, negative_s)))



from bsvpy.modular_inverse import modular_multiplicative_inverse

from bsvpy.ec_point_operation import curve, add, scalar_multiply

from bsvpy.polynomial import Polynomial

from bsvpy.sign_message import message_digest, verify_message

from bsvpy.meta import public_key_to_address

from bsvpy.sign import hash_to_int

import random

from base64 import b64encode





class ThresholdSignature:



    @staticmethod

    def shares_to_points(shares: list) -> list:

        """Returns [(participant_id, share), (participant_id, share), ...]"""

        return [(i + 1, shares[i]) for i in range(len(shares))]



    @staticmethod

    def inspect(items: list) -> str:

        return f'[{", ".join([str(item) for item in items])}]'



    def __init__(self, group_size: int, threshold: int) -> None:

        if group_size < 3:

            raise ValueError(f'Nakasendo group size should be 3 at least')

        self.group_size = group_size

        # t

        self.polynomial_order = threshold - 1

        # t + 1

        self.key_threshold = self.polynomial_order + 1

        # 2t + 1

        self.signature_threshold = 2 * self.polynomial_order + 1

        # Validate t >= 1 and t + 1 <= n and 2t + 1 <= n

        if self.polynomial_order < 1 or self.key_threshold > group_size or self.signature_threshold > group_size:

            raise ValueError(f'Nakasendo threshold should be in interval [2, {(group_size - 1) // 2 + 1}] with {group_size} players')

        # Generate secret shares for each participants

        self.shares, self.public_key = self.jvrss()



    def jvrss(self, debug: bool = False) -> tuple:

        """Returns (shares_of_participants, group_shared_public_key)"""

        if debug:

            print('------------ jvrss ------------')

        # Random polynomials for each player

        polynomials = []

        for i in range(self.group_size):

            p = Polynomial.random(self.polynomial_order, debug)

            if debug:

                print(f'Player {i + 1} {p}')

            polynomials.append(p)

        # Calculate shares for each player

        shares = [0] * self.group_size

        for i in range(self.group_size):

            for j in range(self.group_size):

                fij = polynomials[i].evaluate(j + 1)

                shares[j] += fij

                if debug:

                    print(f'f{i + 1}({j + 1}) = {fij}', end='\t')

            if debug:

                print()

        for i in range(len(shares)):

            shares[i] %= curve.n

        # Calculate group shared public key

        public_key = None

        for i in range(self.group_size):

            public_key = add(public_key, scalar_multiply(polynomials[i].coefficients[0], curve.g))

        if debug:

            secret = sum([p.coefficients[0] for p in polynomials]) % curve.n

            mod_inv_secret = modular_multiplicative_inverse(secret, curve.n)

            print(f'secret = {secret}')

            print(f'mod_inv_secret = {mod_inv_secret}')

            print(f'public key = {public_key}')

            print(f'shares = {ThresholdSignature.inspect(shares)}')

            print('-------------------------------')

        return shares, public_key



    def addss(self, a_shares: list, b_shares: list, debug: bool = False) -> int:

        """Returns secret addition of a and b, with a shares and b shares, without knowing a and b"""

        assert len(a_shares) == self.group_size

        assert len(b_shares) == self.group_size

        if debug:

            print('------------ addss ------------')

            print(ThresholdSignature.inspect(a_shares))

            print(ThresholdSignature.inspect(b_shares))

        shares_addition = [(a_shares[i] + b_shares[i]) % curve.n for i in range(self.group_size)]

        # random pick (t + 1) points

        random_points = random.sample(ThresholdSignature.shares_to_points(shares_addition), self.polynomial_order + 1)

        secrets_addition = Polynomial.interpolate_evaluate(random_points, 0)

        if debug:

            print(f'shares addition = {ThresholdSignature.inspect(shares_addition)}')

            print(f'points picked = {ThresholdSignature.inspect(random_points)}')

            print(f'secrets addition = {secrets_addition}')

            print('-------------------------------')

        return secrets_addition



    def pross(self, a_shares: list, b_shares: list, debug: bool = False) -> int:

        """Returns secret product of a and b, with a shares and b shares, without knowing a and b"""

        assert len(a_shares) == self.group_size

        assert len(b_shares) == self.group_size

        if debug:

            print('------------ pross ------------')

            print(ThresholdSignature.inspect(a_shares))

            print(ThresholdSignature.inspect(b_shares))

        shares_product = [(a_shares[i] * b_shares[i]) % curve.n for i in range(self.group_size)]

        # random pick (2t + 1) points

        random_points = random.sample(ThresholdSignature.shares_to_points(shares_product), 2 * self.polynomial_order + 1)

        secrets_product = Polynomial.interpolate_evaluate(random_points, 0)

        if debug:

            print(f'shares product = {ThresholdSignature.inspect(shares_product)}')

            print(f'points picked = {ThresholdSignature.inspect(random_points)}')

            print(f'secrets product = {secrets_product}')

            print('-------------------------------')

        return secrets_product



    def invss(self, a_shares: list, debug: bool = False) -> list:

        """Returns shares of modular multiplicative inverse of a, with shares of a, without knowing a"""

        assert len(a_shares) == self.group_size

        if debug:

            print('------------ invss ------------')

            print(ThresholdSignature.inspect(a_shares))

        b, _ = self.jvrss(debug)

        u = self.pross(a_shares, b, debug)

        mod_inv_u = modular_multiplicative_inverse(u, curve.n)

        inverse_shares = [(mod_inv_u * bi) % curve.n for bi in b]

        if debug:

            print(f'u = {u}')

            print(f'mod_inv_u = {mod_inv_u}')

            print(f'inverse shares = {ThresholdSignature.inspect(inverse_shares)}')

            random_points = random.sample(ThresholdSignature.shares_to_points(inverse_shares), 2 * self.polynomial_order + 1)

            print(f'points picked = {ThresholdSignature.inspect(random_points)}')

            secret_inverse = Polynomial.interpolate_evaluate(random_points, 0)

            print(f'inverse secret = {secret_inverse}')

            print('-------------------------------')

        return inverse_shares



    def restore_key(self, points: list) -> int:

        """Restore key from given points [(participant_id, share), ...]"""

        if len(points) < self.key_threshold:

            raise ValueError(f'The number of points is less than the threshold')

        return Polynomial.interpolate_evaluate(points, 0)



    def sign_recoverable(self, message: bytes) -> tuple:

        """Create ECDSA compact signature (recovery_id, r, s) with private key shares"""

        e = hash_to_int(message)

        recovery_id, r, s = 0, 0, 0

        while not s:

            mod_inv_k_shares = []

            # Calculate final r

            while not r:

                k_shares, k_public_key = self.jvrss()

                k_x, k_y = k_public_key

                r = k_x % curve.n

                recovery_id = 0 | 2 if k_x > curve.n else 0 | k_y % 2

                mod_inv_k_shares = self.invss(k_shares)

            # Calculate shares of s for each participants

            s_shares = []

            for i in range(self.group_size):

                s_shares.append(((e + r * self.shares[i]) * mod_inv_k_shares[i]) % curve.n)

            # Interpolate shares of s to get final s

            s = Polynomial.interpolate_evaluate(random.sample(ThresholdSignature.shares_to_points(s_shares), self.signature_threshold), 0)

        return recovery_id, r, s



    def sign_message(self, plain_text: str) -> tuple:

        """Sign arbitrary message with private key shares, returns (p2pkh_address, serialized_compact_signature)"""

        d = message_digest(plain_text)

        # recovery signature

        recovery_id, r, s = self.sign_recoverable(d)

        # prefix = 27 + recovery_id + (4 if using compressed public key else 0)

        prefix = 27 + recovery_id + 4

        serialized_sig = prefix.to_bytes(1, byteorder='big') + r.to_bytes(32, byteorder='big') + s.to_bytes(32, byteorder='big')

        return public_key_to_address(self.public_key, compressed=True), b64encode(serialized_sig).decode('ascii')





if __name__ == '__main__':

    ts = ThresholdSignature(group_size=3, threshold=2)

    # Plain text to sign

    plain = 'Threshold Signature Scheme Sign Test\nPrivate key shares:\n' + ThresholdSignature.inspect(ts.shares)

    print(plain)

    print('------------------')

    # Sign message

    address, sig = ts.sign_message(plain)

    print(address, sig)

    print('------------------')

    # Verify signature

    print(verify_message(address, plain, sig))



def extended_euclid_gcd(a: int, b: int) -> list:
    """
    Returns [gcd(a, b), x, y] where ax + by = gcd(a, b)
    """
    s, old_s = 0, 1
    t, old_t = 1, 0
    r, old_r = b, a
    while r != 0:
        quotient = old_r // r
        old_r, r = r, old_r - quotient * r
        old_s, s = s, old_s - quotient * s
        old_t, t = t, old_t - quotient * t
    return [old_r, old_s, old_t]


def modular_multiplicative_inverse(a: int, n: int) -> int:
    """
    Assumes that a and n are co-prime, returns modular multiplicative inverse of a under n
    """
    # Find gcd using Extended Euclid's Algorithm
    gcd, x, y = extended_euclid_gcd(a, n)
    # In case x is negative, we handle it by adding extra n
    # Because we know that modular multiplicative inverse of a in range n lies in the range [0, n-1]
    if x < 0:
        x += n
    return x


if __name__ == '__main__':
    print(modular_multiplicative_inverse(3, 7))